/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "minrtsFilterUtil.h"

minrtsFilter *minrtsMakeFilter (int nData, float *spFreq, float *trans) {
/* ******************************************************************
   * make a filter object from explicit supplied data
   * Inputs:
   *          nData  & number of input points
   *          spFreq & input spatial Freq (meters^-1)
   *          trans  & filter transmission
   ******************************************************************* */
   int i;
   long *ind;
   minrtsFilter *filt;
   filt = (minrtsFilter *)malloc(sizeof(minrtsFilter));
   filt->nData = nData;
   filt->spFreq = (float *) malloc (nData * sizeof(float));
   filt->trans  = (float *) malloc (nData * sizeof(float));
/*   
      Spline interpolation requires that frequencies be sorted in
      increasing order 
*/
   ind = minrtsHeapSortI(nData, spFreq );
   for (i=0; i<nData; i++) {
      filt->spFreq[i]  = spFreq[i]; /* spFreq is already sorted */
      filt->trans[i]   = trans[ind[i]];
   }
   free(ind);
   return filt;
} /* end of minrtsMakeFilter */

minrtsFilter *minrtsMakeFilterFromFile (char *fileName, int *ierr) {
/* ******************************************************************
   * make a filter object based on data in a ASCII file
   * 1st line of file should specify type: REAL, COMPLEX or AMPPHASE
   * Next line with number of values
   * following lines have wavelength (microns) and transmission
   * Input wavelengths should be in strictly ascending order
   * Inputs:
   *           fileName   & name of disk file
   *           ierr       & 0 is ok
   ****************************************************************** */
   int nData, i, iFile, iType;
   float micron, trans, trans2;
   char type[10];
   float twopi, rad;
   FILE *filterFile;
   minrtsFilter *filter = NULL;
   long *ind; /* for sorting */
   float *tTrans=NULL, *tFreq=NULL; /* for sorting */
   twopi = 3.1415926535*2.;
   rad = twopi/360.;
   *ierr = 0;
   filterFile = fopen(fileName, "r");
   if (filterFile == NULL) {
      *ierr = errno;
      printf ("couldnt open filter file %s \n", fileName);
      return NULL;
   }
   fscanf(filterFile," %s", type);
   if (strcmp(type, "REAL") == 0) iType = 0;
   else if (strcmp(type, "COMPLEX") == 0) iType = 1;
   else if (strcmp(type, "AMPPHASE") == 0) iType = 2;
   else { /* assume REAL if not specified, but then rewind */
      iType=0;
      fclose(filterFile);
      filterFile = fopen(fileName, "r");
   }
   fscanf(filterFile, " %i", &nData);
   if (nData >0) {
      filter = (minrtsFilter *)malloc(sizeof(minrtsFilter));
      filter->nData = nData;
      tFreq = (float *) malloc (nData * sizeof(float));
      filter->spFreq = (float *) malloc (nData * sizeof(float));
      if (iType == 0) {
         tTrans  = (float *) malloc (nData * sizeof(float));
         filter->trans  = (float *) malloc (nData * sizeof(float));
      } else {
         tTrans  = (float *) malloc (nData * 2 * sizeof(float));
         filter->trans = (float *)malloc (nData * 2 * sizeof(float));
      }
   }
   for (i=0; i<nData; i++) {
      if (iType==0) {
         iFile = fscanf(filterFile, " %f %f", &micron, &trans); 
         if (iFile != 2) {
            *ierr = errno;
            printf ("read error %i filter file %s \n", errno, fileName);
            minrtsKillFilter (filter);
            free(tTrans);
            free(tFreq);
            return NULL;
         }
      } else {
         iFile = fscanf(filterFile, " %f %f %f", &micron, &trans, &trans2); 
         if (iFile != 3) {
            *ierr = errno;
            printf ("read error %i filter file %s \n", errno, fileName);
            minrtsKillFilter (filter);
            free(tTrans);
            free(tFreq);
            return NULL;
         }
      }
      if (iType==0) tTrans[i] = trans;
      else if (iType==1) {
         tTrans[2*i]   = trans;
         tTrans[2*i+1] = trans2;
      } else if (iType == 2) {
         tTrans[2*i]   = trans* cos(rad * trans2);
         tTrans[2*i+1] = trans * sin(rad * trans2);
      }  
      tFreq[i] = twopi*1.e6/micron; /* wavelength in microns -> waveno m^-1 */
   }
   fclose (filterFile);
   filter->type = iType;
   if (iType == 2) filter->type=1; /* we have already converted ampphase->cossin */
/* sort ascending for later interpolation */
   ind = minrtsHeapSortI(nData, tFreq );
   for (i=0; i<nData; i++) {
      filter->spFreq[i] = tFreq[i]; /* tFreq is already sorted */
      if (iType==0) filter->trans[i] = tTrans[ind[i]];
      else {
         filter->trans[2*i]   = tTrans[2*ind[i]];
         filter->trans[2*i+1] = tTrans[2*ind[i]+1];
      }
   }
   free(ind);
   free(tFreq);
   free(tTrans);
   return filter;
} /* end of Make Filter From File */
void minrtsKillFilter ( minrtsFilter* filt ){
/* ******************************************************************
   * destroy memory of stored values in filter object
   * filt   & filter object
   ***************************************************************** */
   free (filt->spFreq);
   free (filt->trans);
   free (filt);
   return;
} /* end of minrtsKillFilter */

float *minrtsSpline(int nX, float *x, float *y, int nT, float *t, float sigma){
/* ******************************************************************
   * general utility for cubic spline interpolation
   * Inputs:
   *          nX  & number of input tabular points
   *          x   & tabular abscissa values; must be monotonically increasing
   *          y   & tabular oordinate values
   *         nT   & number of desired output values
   *          t   & abscissa values of desired output; must be monotonically increasing
   *          sigma & tightness parameter  <<1 = spline >>1 = polynomial
   * Returns:
   *              & interpolated values
   *********************************************************************** */
   long i;
   float *yp, *spl;    /* temp storage */
   float delx1, dx1, delx2, delx12, dx2 = 0., c1, c2, c3;
   float slpp1, deln, delnm1, delnn, slppn;
   float s;
   long nm1, np1, j;
   long *subs, su, s1;
   int crazy;
   float sigmap, del1, del2, dels, exps, 
      exps1, sinhs, sinhd1, sinhd2, sinhin, diag1, diag2, diagin, spdiag;
   yp = (float *) malloc (2*nX*sizeof(float));
   spl = (float *) malloc (nT*sizeof(float));
   subs = (long *) malloc (nT*sizeof(long));
   delx1 = x[1] - x[0];
   dx1 = (y[1]-y[0])/delx1;
   nm1 = nX - 1;
   np1 = nX + 1;
   delx2 = x[2] - x[1];
   delx12 = x[2] - x[0];
   c1 = -(delx12+delx1)/delx12/delx1;
   c2 = delx12/delx1/delx2;
   c3 = -delx1/delx12/delx2;
   slpp1 = c1*y[0]+c2*y[1]+c3*y[2];
   deln = x[nm1] - x[nm1-1];
   delnm1 = x[nm1-1] - x[nm1-2];
   delnn = x[nm1] - x[nm1-2];
   c1 = (delnn+deln)/delnn/deln;
   c2 = -delnn/deln/delnm1;
   c3 = deln/delnn/delnm1;
   slppn = c3*y[nm1-2]+c2*y[nm1-1]+c1*y[nm1];

   sigmap = sigma*nm1/(x[nm1]-x[0]);
   dels = sigmap*delx1;
   exps = exp(dels);
   sinhs = 0.5*(exps-1./exps);
   sinhin = 1./(delx1*sinhs);
   diag1 = sinhin*(dels*.5*(exps+1./exps)-sinhs);
   diagin = 1./diag1;
   yp[0] = diagin*(dx1-slpp1);
   spdiag = sinhin*(sinhs-dels);
   yp[nX] = diagin*spdiag;
   for (i=1;i<= nm1-1;i++) {
      delx2 = x[i+1] - x[i];
      dx2 = (y[i+1]-y[i])/delx2;
      dels = sigmap*delx2;
      exps = exp(dels);
      sinhs = .5*(exps-1./exps);
      sinhin=1./(delx2*sinhs);
      diag2=sinhin*(dels*(.5*(exps+1./exps))-sinhs);
      diagin = 1./(diag1+diag2-spdiag*yp[nX+i-1]);
      yp[i] = diagin*(dx2-dx1-spdiag*yp[i-1]);
      spdiag=sinhin*(sinhs-dels);
      yp[i+nX]=diagin*spdiag;
      dx1=dx2;
      diag1=diag2;
   }
   diagin=1./(diag1-spdiag*yp[nX+nm1-1]);
   yp[nm1]=diagin*(slppn-dx2-spdiag*yp[nm1-1]);
   for (i=nX-2;i>=0;i--) yp[i]=yp[i]-yp[i+nX]*yp[i+1];
   for (i=0;i<nT;i++) subs[i] = nm1;
   s = x[nm1] - x[0];
   sigmap = sigma*nm1/s;
   j = 0;
   for (i=1; i<=nm1; i++) {
      while (x[i] > t[j]) {
         subs[j] = i;
         j++;
         if (j == nT) goto done;
      }
   }
   done:
   for (i=0;i<nT;i++) {
      if (t[i] < x[0]) spl[i] = 0.;
      else if (t[i] > x[nX-1]) spl[i] = 0.;
      else {
         su = subs[i];
         s1 = su-1;
         del1 = t[i] - x[s1];
         del2 = x[su] - t[i];
         dels = x[su] -x[s1];
         exps1 = exp(sigmap*del1);
         crazy = (exps1 < 1.e-8 || exps1 > 1.e8);
         sinhd1 = .5*(exps1-1./exps1);
         exps = exp(sigmap*del2);
         crazy = (crazy || exps < 1.e-8 || exps > 1.e8);
         sinhd2 = .5*(exps - 1./exps);
         exps = exps1 * exps;
         sinhs = .5*(exps - 1./exps);
         if (crazy) spl[i] = 0.; else 
            spl[i] = (yp[su]*sinhd1 + yp[s1]*sinhd2)/sinhs +
               ((y[su]-yp[su])*del1 + (y[s1]-yp[s1])*del2)/dels;
      }
   }
   free(yp);
   free(subs);
   return spl;
} /* end of minrtsSpline */

float *minrtsFilterInterp (minrtsFilter *filt, int nWave, float *spFreq) {
/* ******************************************************************
   * spline interpolation of stored filter description onto given spatial freq array
   * For COMPLEX stored values, only REAL part is returned;
   * For negative values of spFreq, trans of equivalent positive value is returned
   * values for spFreq outside stored table are zeroed. 
   * Inputs:
   *          filt     &  filter object
   *          nWave    &  number of desired interpolation points
   *          spFreq   &  array of spatial Frequencies (meter^-1)
   * Returns:
   *                   & array of interpolated values
   ***************************************************************************** */
   int i;
   float minwave, maxwave;
   int xmax, xmin;
   float *outTrans, *t, *realTrans, *sortWave;
   long *waveIndex;
   float sf;
   outTrans  = (float *)malloc(nWave * sizeof(float)); /* real */
   sortWave  = (float *)malloc(nWave * sizeof(float)); /* copy for sorting */
   realTrans = (float *)malloc(filt->nData * sizeof(float));
   if (filt->type == 0) for (i=0;i<filt->nData;i++) realTrans[i] = filt->trans[i];
      else for (i=0;i<filt->nData;i++) realTrans[i] = filt->trans[2*i];
   minrtsMinMax(filt->nData, filt->spFreq, &minwave, &maxwave, &xmin, &xmax);
   for (i=0;i<nWave;i++) /* copy abs value for sorting */
      if (spFreq[i] > 0.) sortWave[i] = spFreq[i]; 
         else sortWave[i] = -spFreq[i];
   waveIndex = minrtsHeapSortI(nWave, sortWave );
   t = minrtsSpline(filt->nData, filt->spFreq, realTrans, nWave, sortWave, 1.);
   for (i=0;i<nWave;i++) {
      sf = sortWave[i];
      if ((sf >= maxwave) || (sf <=minwave)) t[i] = 0.; /* zero out of range */
      outTrans[waveIndex[i]] = t[i];
   }
   free(t);
   free(sortWave);
   free(waveIndex);
   free(realTrans);
   return outTrans;
} /* end of minrtsFilterInterp */


float *minrtsFilterComplexFilter (minrtsFilter *filt, int nWave, float *spFreq) {
/* ******************************************************************
   * spline interpolation of stored filter description onto given spatial freq array
   * values for spFreq to produce COMPLEX output filter.
   * For negative wavenumber the complex conjugate of the stored value is returned.
   * Inputs:
   *          filt     &  filter object
   *          nWave    &  number of desired interpolation points; must be increasing
   *          spFreq   &  array of spatial Frequencies (meter^-1)
   * Returns:
   *                   & array of interpolated values as complex
   ***************************************************************************** */
   int i;
   float minwave, maxwave;
   int xmax, xmin;
   float *outTrans, *reTrans, *imTrans, *re, *im, sf;
   float *sortWave;
   long *waveIndex;
   outTrans = (float *)malloc(2 * nWave * sizeof(float)); /* complex */
   reTrans = (float *)malloc(filt->nData * sizeof(float)); /* complex */
   imTrans = (float *)malloc(filt->nData * sizeof(float)); /* complex */
   sortWave  = (float *)malloc(nWave * sizeof(float)); /* copy for sorting */
   if (filt->type == 0) {
      for (i=0;i<filt->nData;i++) {
         reTrans[i] = filt->trans[i];
         imTrans[i] = 0.;
      }
   } else {
      for (i=0;i<filt->nData;i++) {
         reTrans[i] = filt->trans[2*i];
         imTrans[i] = filt->trans[2*i+1];
      }
   }
   minrtsMinMax(filt->nData, filt->spFreq, &minwave, &maxwave, &xmin, &xmax);
   for (i=0;i<nWave;i++) /* copy abs value for sorting */
      if (spFreq[i] > 0.) sortWave[i] = spFreq[i]; 
         else sortWave[i] = -spFreq[i];
   waveIndex = minrtsHeapSortI(nWave, sortWave);
   re = minrtsSpline(filt->nData, filt->spFreq, reTrans, nWave, sortWave, 1.);
   im = minrtsSpline(filt->nData, filt->spFreq, imTrans, nWave, sortWave, 1.);
   for (i=0;i<nWave;i++) {
      sf = sortWave[i];
      if ((sf >= maxwave) || (sf <=minwave)) {/* zero out of range */
         re[i] = 0.;
         im[i] = 0.;
      }
      outTrans[2*waveIndex[i]]   = re[i];
      if(spFreq[waveIndex[i]] > 0.) outTrans[2*waveIndex[i]+1] = im[i];
         else outTrans[2*waveIndex[i]+1] = -im[i];
   }
   free(re); 
   free(im);
   free(reTrans);
   free(imTrans);
   free(sortWave);
   free(waveIndex);
   return outTrans;
} /* end of minrtsFilterComplexFilter */

void minrtsMoments(int nData, float *data, float *mean, float *rms){
/* ******************************************************************
   * general utility to evalue mean and rms of a vector
   * Inputs:
   *          nData  & number of points in vector
   *          data   & data
   * Outputs:
   *          mean    & mean value of array
   *          rms     & rms value of array
   ****************************************************************** */
   float sum, sum2, x;
   int i;
   sum = 0;
   sum2 = 0;
   for (i=0; i<nData; i++){
      x = data[i];
      sum = sum+x;
      sum2 = sum2+ x*x;
   }
   sum = sum/nData;
   *rms = sqrt(sum2/nData - sum*sum);
   *mean = sum;
   return;
}

void minrtsMinMax(int nData, float *data, float *min, float *max,
   int *minPos, int *maxPos){
/* ******************************************************************
   * general utility to find min and max (and positions) of a vector
   * Inputs:
   *          nData  & number of points in vector
   *          data   & data
   * Outputs:
   *          min    & min value of vector
   *          max    & max value of vector
   *       minPos    & pos of 1st occurance of min value (0-rel)
   *       maxPos    & pos of 1st occurance of max value
   ****************************************************************** */
   int i, map, mip;
   float mx, mn, x;
   mx = data[0];
   mn = mx;
   map = 0;
   mip = 0;
   for (i=1; i< nData; i++) {
      x = data[i];
      if(x > mx) {mx = x; map = i;}
      if(x < mn) {mn = x; mip = i;}
   }
   *max = mx;
   *min = mn;
   *maxPos = map;
   *minPos = mip;
   return;
}  /* end min max */

void minrtsAbs(int nData, float *data) {
/* ******************************************************************
   * general utility to take (in place) abs value of a vector
   * Inputs:
   *          nData  & number of points in vector
   *          data   & data
   * Outputs:
   *          data    & absolute value of input
   ***************************************************************** */
   int i;
   for(i=0; i<nData; i++) data[i] = fabs(data[i]);
   return;
}

void minrtsFilterWidth(minrtsFilter *filt, float *mean, float *width){
  /* ******************************************************************
   * compute mean and rms widtth of stored filter 
   * Inputs:
   *          filt     &  filter object
   * Outputs:
   *          mean     &  average of filter sp freq (meters^-1)
   *          width    &  rms width of filter (meters^-1)
   ******************************************************************** */
   int i;
   float sum, sum2, w, t;
   sum = 0;
   sum2 = 0;
   for (i=0; i<filt->nData; i++) {
      t = filt->trans[i];
      w = filt->spFreq[i];
      sum = sum + t*w;
      sum2= sum2+ t*w*w;
   }
   sum = sum/filt->nData;
   sum2 = sqrt(sum2/filt->nData - sum*sum);
   *mean = sum;
   *width = sum2;
   return;
}  /*  end of minrtsFilterWidth */

void minrtsBoxSmooth(long n, float *inData, float *outData, int boxSize) {
  /* ******************************************************************
   * run a boxcar smoothing over input data;
   * 
   * Inputs:
   *          n       &  number of points in input data
   *          inData  & input data
   *          boxSize & size of box in pixels
   *          outData & preallocated array for output data
   * Outputs:
   *          outData & smoothed version of indata; memory for this must be preallocated
   ******************************************************************** */
   int i,j;
   float fj;
   float sum, fBox, fBox2;

   fBox  = (float)boxSize;
   fBox2 = 0.5 * fBox;
   sum = 0.;
   fj  = -.5;
   for (i=0;i<boxSize; i++) {
      sum = sum + inData[i];
      fj = 0.5 + fj;
      j  = (int)(fj + .5);
      outData[j] = sum/((float)(i+1));
   }
   for (i=boxSize; i<n; i++) {
      sum = sum + inData[i] - inData[i-boxSize];
      fj = 1. + fj;
      j = (int)(fj + .5);
      outData[j] = sum/fBox;
   }
   for (i=n;i<(boxSize+n-1); i++) {
      fj = fj + .5;
      sum = sum - inData[i-boxSize];
      j = (int)(fj + .5);
      outData[j] = sum/((float)(boxSize+n-i-1));
   }
   sum = 0.;
   return ;
}
void minrtsHeapSort(long n, float *ra)
{
   long i,ir,j,l;
   float rra;

   if (n < 2) return;
   l=(n >> 1)+1;
   ir=n;
   for (;;) {
      if (l > 1)  rra=ra[--l-1];  
      else {
         rra=ra[ir-1];
         ra[ir-1]=ra[1-1];
         if (--ir == 1) {
            ra[0]=rra;
            break;
         }
      }
      i=l;
      j=l+l;
      while (j <= ir) {
         if (j < ir && ra[j-1] < ra[j]) j++;
         if (rra < ra[j-1]) {
            ra[i-1]=ra[j-1];
            i=j;
            j <<= 1;
         } else j=ir+1;
      }
      ra[i-1]=rra;
   }
}


long *minrtsHeapSortI(long n, float *ra)
{
   long i,ir,j,l,ri;
   long *ind;
   float rra;

   ind = (long *)malloc(n * sizeof(long));
   for (i=0; i< n; i++) ind[i] = i;
   if (n < 2) return ind;
   l=(n >> 1)+1;
   ir=n;
   for (;;) {
      if (l > 1) {
         rra=ra[--l-1];
         ri = ind[l-1];
      } else {
         rra=ra[ir-1];
         ri = ind[ir-1];
         ra[ir-1]=ra[1-1];
         ind[ir-1] = ind[l-1];
         if (--ir == 1) {
            ra[0]=rra;
            ind[0] = ri;
            break;
         }
      }
      i=l;
      j=l+l;
      while (j <= ir) {
         if (j < ir && ra[j-1] < ra[j]) j++;
         if (rra < ra[j-1]) {
            ra[i-1]=ra[j-1];
            ind[i-1] = ind[j-1];
            i=j;
            j <<= 1;
         } else j=ir+1;
      }
      ra[i-1]=rra;
      ind[i-1] = ri;
   }
return ind;
}
